#define G_LOG_DOMAIN "boilerplate-application"

#include "config.h"

#include "boilerplate-application.h"
#include "boilerplate-window.h"
#include "boilerplate-preferences.h"
#include "boilerplate-plugins-engine.h"
#include "boilerplate-activatable.h"

#include <boilerplate.h>

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <libpeas/peas.h>

/**
 * SECTION:boilerplate-application
 * @title: BoilerplateApplication
 * @short_description: The current application
 * @include: boilerplate-application.h
 * @see_also: #GtkApplication
 *
 * The #BoilerplateApplication which connects the different parts of the
 * application. Also contains some helper functions for other parts of the
 * application.
 **/

typedef struct _BoilerplateApplicationPrivate
{
    GSettings *settings;

    BoilerplatePluginsEngine *plugins_engine;
    PeasExtensionSet *extensions;

    gboolean dark_mode;
} BoilerplateApplicationPrivate;

struct _BoilerplateApplication
{
    GtkApplication parent_instance;

    BoilerplateWindow *window;
};

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateApplication, boilerplate_application, GTK_TYPE_APPLICATION)

static const GOptionEntry option_entries[] = {
    { "version", '\0', 0, G_OPTION_ARG_NONE, NULL, "Print release version and exit", NULL },
    { "dark-mode", '\0', 0, G_OPTION_ARG_NONE, NULL, "Enable dark theme", NULL },
    //{ "traditional-mode", 'c', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, NULL, "Disable Client Side Decorations (CSD)", NULL },
    { NULL }
};

/**
 * boilerplate_application_set_accelerator:
 * @application: The application (self)
 * @action_name: The name of the action, eg. app.preferences, win.show-help-overlay
 * @accel: The accelerator, eg. Primary+N
 *
 * A helper function to quickly create an single accelerator
 *
 * ## Usage
 * |[
 * boilerplate_application_set_accelerator (GTK_APPLICATION (application),
 *                                          "app.help",
 *                                          "F1");
 * ]|
 */
void
boilerplate_application_set_accelerator (GtkApplication *application,
                                         const gchar    *action_name,
                                         const gchar    *accel)
{
    const gchar *vaccels[] = {
        accel,
        NULL
    };

    gtk_application_set_accels_for_action (application, action_name, vaccels);
}

/**
 * boilerplate_application_set_accelerators:
 * @application: A #GtkApplication instance
 * @action_name: The name of the action, eg. `app.preferences` or `win.show-help-overlay`
 * @accels: An array of accelerator to add under the @action_name
 *
 * A helper function to quickly create multiple accelerators for one @action_name
 *
 * ## Usage
 * |[
 * const gchar *help_accels[] = {
 *     "F1",
 *     "<Primary>?",
 *     NULL
 * };
 *
 * boilerplate_application_set_accelerators (GTK_APPLICATION (application),
 *                                           "app.help",
 *                                           help_accels);
 * ]|
 */
void
boilerplate_application_set_accelerators (GtkApplication *application,
                                          const gchar    *action_name,
                                          const gchar   **accels)
{
    gtk_application_set_accels_for_action (application, action_name, accels);
}

/**
 * boilerplate_application_show_uri:
 * @user_data: A #GApplication instance
 * @uri: A #gchar string that contains the url to open
 *
 * A helper function for gtk_show_uri_on_window(), this function takes a
 * #GApplication instance as @user_data and a #gchar string as @uri to show.
 * This function loads the proper screen, and checks to see if errors occure.
 * If so they are shown to the user as a #GtkDialog instead of in the terminal.
 *
 * ## Usage
 * |[
 * boilerplate_application_show_uri (G_APPLICATION (application),
 *                                   "https://www.gnome.org");
 * ]|
 */
void
boilerplate_application_show_uri (GApplication *user_data,
                                  const gchar  *uri)
{
    g_debug ("Opening URI window for: %s", uri);

    GtkApplication *application;
    GtkWindow *window;
    GtkWidget *dialog;
    GError *error = NULL;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

    gtk_show_uri_on_window (window,
                            uri,
                            gtk_get_current_event_time (),
                            &error);

    if (error)
    {
        dialog = gtk_message_dialog_new (window ? GTK_WINDOW (window) : NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_OK,
                                         _("There was an error displaying help: \n%s"),
                                         error->message);

        g_signal_connect (G_OBJECT (dialog), "response",
                          G_CALLBACK (gtk_widget_destroy),
                          NULL);

        gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
        gtk_widget_show (dialog);
        g_error_free (error);
    }
}

/*
 * on_preferences:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * Opens the application's preferences.
 */
static void
on_preferences (GSimpleAction *action,
                GVariant      *parameter,
                gpointer       user_data)
{
    g_debug ("Launching preferences");

    GtkApplication *application;
    BoilerplateWindow *window;
    BoilerplatePreferences *preferences;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    preferences = boilerplate_preferences_new (window);

    gtk_window_present (GTK_WINDOW (preferences));
}

/*
 * on_help:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * Opens the application's help window.
 */
static void
on_help (GSimpleAction *action,
         GVariant      *parameter,
         gpointer       user_data)
{
    g_debug ("Launching help");

    boilerplate_application_show_uri (G_APPLICATION (user_data),
                                      "help:" PACKAGE);
}

/*
 * on_about:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * Opens the application's about dialog.
 */
static void
on_about (GSimpleAction *action,
          GVariant      *parameter,
          gpointer       user_data)
{
    g_debug ("Launching about");

    GtkApplication *application;
    GtkWindow *window;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

	static const char *authors[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const char *artists[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const gchar *translators =
        "Julian Richen <julian@richen.io>";

	gtk_show_about_dialog (window,
		                   "program-name", g_get_application_name (),
		                   "version", PACKAGE_VERSION,
		                   "copyright", PACKAGE_COPYRIGHT,
		                   "comments", PACKAGE_SUMMARY,
		                   "authors", authors,
		                   "artists", artists,
		                   "translator-credits", translators,
		                   "website-label", "Website",
		                   "website", PACKAGE_URL_HOMEPAGE,
		                   "logo-icon-name", gtk_window_get_default_icon_name (),
		                   "wrap-license", TRUE,
		                   "license-type", GTK_LICENSE_GPL_3_0,
		                   NULL);
}

/*
 * on_quit:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * Action to quit the application.
 */
static void
on_quit (GSimpleAction *action,
         GVariant      *parameter,
         gpointer       user_data)
{
    g_debug ("Quiting application");

    GtkApplication *application = GTK_APPLICATION (user_data);
    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);

    gtk_widget_destroy (GTK_WIDGET (self->window));

    g_application_quit (G_APPLICATION (user_data));
}

/*
 * action_entries[]
 *
 * Connects the the app-menu and menubar links in the respective menu to an
 * action. The "Keyboard Shortcuts" window is known as the help-overlay internally.
 * It is not listed here because the GtkShortcutsWindow with an id of help_overlay
 * is included as a resource in gresource via gtk/help-overlay.ui. Gtk will
 * automaticlly include and connect it with an action of win.show-help-overlay
 * and default shortcuts (Control-F1 and Control-?).
 */
static GActionEntry action_entries[] = {
    { "preferences", on_preferences, NULL, NULL, NULL },
    { "help", on_help, NULL, NULL, NULL },
    { "about", on_about, NULL, NULL, NULL },
	{ "quit", on_quit,  NULL, NULL, NULL },
};

/*
 * show_menubar:
 *
 * Determines from gtk_settings_get_default() if the current #GtkApplication
 * should show it's menubar via the `gtk-shell-shows-menubar` key. This key is
 * set by the system packagers, as some distributions may want them on, even if
 * they break Client Side Decorations (CSD).
 *
 * Returns: #gboolean; %TRUE to show menubar and %FALSE to not show menubar
 */
static gboolean
show_menubar (void)
{
	GtkSettings *settings = gtk_settings_get_default ();
	gboolean result;

	g_object_get (settings,
	              "gtk-shell-shows-menubar", &result,
	              NULL);

	return result;
}

/*
 * boilerplate_application_shutdown:
 * @application: A #GApplication instance
 *
 * Shutdown the application, all clean-up should be done here. I.e saving data,
 * states, etc...
 */
/*
 * Application Sequence: Final
 */
static void
boilerplate_application_shutdown (GApplication *application)
{
    g_debug ("Application shutting down");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    g_object_unref (priv->plugins_engine);

    G_APPLICATION_CLASS (boilerplate_application_parent_class)->shutdown (application);
}

/*
 * boilerplate_application_activate:
 * @application: A #GApplication instance
 *
 * Activates the #BoilerplateWindow, a sub-object of #GtkApplicationWindow. It
 * then presents the window to the user.
 */
/*
 * Application Sequence: 04
 */
static void
boilerplate_application_activate (GApplication *application)
{
    g_debug ("Application activated");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);

    /*
     * Create GtkApplicationWindow if needed
     * If you need multiple windows open (like you have multiple instances) you
     * should convert self->window to a GList *windows and loop the list to
     * select the window
     */
    if (self->window == NULL) {
	    self->window = boilerplate_window_new (BOILERPLATE_APPLICATION (application));

        g_debug ("GtkApplicaitonWindow created and connected");
    }

    g_debug ("Presenting GtkApplicaitonWindow");

	gtk_window_present (GTK_WINDOW (self->window));
}

/*
 * boilerplate_application_startup:
 * @application: A #GApplication instance
 *
 * Starts many of the applications features when the application is first
 * created. It also sets information used throughout the application like
 * the application:
 *  * name
 *  * icon
 *  * action entries
 *  * accelerators
 *  * menubar (fallback)
 *  * plugins
 */
/*
 * Application Sequence: 03
 */
static void
boilerplate_application_startup (GApplication *application)
{
    g_debug ("Application startup");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    /* Starting-up application */
	G_APPLICATION_CLASS (boilerplate_application_parent_class)->startup (application);

    /* Init menus */
    g_debug ("Creating application action entries");

    /* Add action callbacks */
	g_action_map_add_action_entries (G_ACTION_MAP (application),
                                     action_entries, G_N_ELEMENTS (action_entries),
                                     application);

    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.preferences", "<Primary>comma");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.help", "F1");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.about", "F2");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.quit", "<Primary>q");

    /* Disable fallback menubar by default */
    if (!show_menubar ()) {
        gtk_application_set_menubar (GTK_APPLICATION (application), NULL);
    }

    priv->plugins_engine = boilerplate_plugins_engine_get_default ();

    /* Default settings */
    priv->dark_mode = FALSE;
}

/*
 * boilerplate_application_handle_local_command_line:
 * @application:  A #GApplication instance
 * @options: A #GVariantDict which normally contains a #GOptionEntry of command-
 * line arguments.
 *
 * Handles the commandline flags like `--version` that do not have a variable
 * set as a pointer in #GOptionEntry's `option_entries[]`.
 *
 * So if the entry was like so:
 * |[
 * { "example", 'e', 0, G_OPTION_ARG_NONE, &EXAMPLE, "example desc", NULL },
 * ]|
 *
 * It would set the value of the `--example`/`-e` flag into the `EXAMPLE` constant
 *
 * But if the entry was like so:
 * |[
 * { "example", 'e', 0, G_OPTION_ARG_NONE, NULL, "example desc", NULL },
 * ]|
 *
 * It will be passed into the `handle_local_command_line()` function via the
 * ::handle_local_options virtual functioin. Were it can be accessed with
 * g_variant_dict_lookup():
 *
 * |[
 * g_variant_dict_lookup (options, "example", "b", &priv->example);
 * ]|
 *
 * Returns: #gint; `-1` continue or `0` proccess something
 */
/*
 * Application Sequence: 02
 */
static gint
boilerplate_application_handle_local_command_line (GApplication *application,
                                                   GVariantDict *options)
{
    g_debug ("Application handling local command line");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    g_variant_dict_lookup (options, "dark-mode", "b", &priv->dark_mode);

    if (g_variant_dict_contains (options, "version"))
    {
        g_print ("%s\n", PACKAGE_VERSION);
        return 0;
    }

    return -1;
}

/*
 * boilerplate_application_init:
 * @self: An instance of #BoilerplateApplication
 *
 * Initializes various aspects of the application. Setting the g_set_application_name(),
 * gtk_window_set_default_icon_name(), g_application_add_main_option_entries(),
 * and so on should be done here. Anything the application would need before
 * becoming registered in the #GMainLoop.
 */
static void
boilerplate_application_init (BoilerplateApplication *self)
{
    //BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    g_debug ("Application init");

    /* Set application name */
	g_set_application_name (PACKAGE_NAME);

    g_debug ("Application name: %s", g_get_application_name ());

    /* Set default icon via name */
	gtk_window_set_default_icon_name (PACKAGE_ID);

    g_debug ("Application icon name: %s", gtk_window_get_default_icon_name ());

    /* Add option entries / command line flags */
	g_application_add_main_option_entries (G_APPLICATION (self), option_entries);

    /* Connect settings */
    g_debug ("Loading application settings");
    //priv->settings = g_settings_new (PACKAGE_ID);
}

/*
 * boilerplate_application_class_init:
 * @klass: An instance of #BoilerplateApplicationClass
 *
 * Initializes the various virtural functions in #BoilerplateApplicationClass.
 * The #BoilerplateApplicationClass is created automatically thanks to the
 * #G_DEFINE_TYPE_WITH_PRIVATE macro.
 */
static void
boilerplate_application_class_init (BoilerplateApplicationClass *klass)
{
    GApplicationClass *application_class = G_APPLICATION_CLASS (klass);

	application_class->startup = boilerplate_application_startup;
	application_class->activate = boilerplate_application_activate;
    application_class->shutdown = boilerplate_application_shutdown;
    application_class->handle_local_options = boilerplate_application_handle_local_command_line;
}

/**
 * boilerplate_application_new:
 *
 * Creates a new #BoilerplateApplication, a sub-object of #GtkApplication.
 *
 * ## Usage
 * Generally the #BoilerplateApplication is called in the main entry point for
 * the executable. In most cases this is `main.c`.
 *
 * |[
 * int
 * main (int    argc,
 *       char **argv)
 * {
 *     BoilerplateApplication *application;
 *     int status;
 *
 *     application = boilerplate_application_new ();
 *     g_application_set_default (G_APPLICATION (application));
 *
 *     status = g_application_run (G_APPLICATION (application), argc, argv);
 *
 *     g_object_unref (application);
 *
 *     return status;
 * }
 * ]|
 *
 * Returns: #BoilerplateApplication; a sub-object of #GtkApplication
 */
/*
 * Application Sequence: 01
 */
BoilerplateApplication *
boilerplate_application_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_APPLICATION,
                         "application-id", PACKAGE_ID,
                         "flags", G_APPLICATION_FLAGS_NONE,
                         "resource-base-path", PACKAGE_RESOURCES,
                         NULL);
}

